from django.db import models
from django.test import TestCase

from fields import MeasurementField
import measure
import utils

class MeasurementTestModel(models.Model):
    measurement = MeasurementField()

    def __unicode__(self):
        return unicode(self.pk)

class MeasurementFieldTest(TestCase):
    def test_storage_and_retrieval_of_measurement(self):
        arbitrary_default_unit='lb'
        arbitrary_value=124
        arbitrary_measure=measure.Weight
        arbitrary_measurement=utils.get_measurement(
            arbitrary_measure,
            arbitrary_value,
            arbitrary_default_unit,
        )

        instance = MeasurementTestModel.objects.create(
            measurement=arbitrary_measurement
        )
        instance.save()

        retrieved = MeasurementTestModel.objects.all()[0]

        self.assertEqual(
            retrieved.measurement._default_unit,
            arbitrary_default_unit,
        )
        self.assertEqual(
            retrieved.measurement.__class__,
            arbitrary_measure,
        )
        self.assertAlmostEqual(
            retrieved.measurement.lb,
            arbitrary_value,
        )

    def test_retrieval_of_unknown_measure_returns_unknown_measure(self):
        arbitrary_default_unit='versta'
        arbitrary_measure_name='ImperialRussian'
        arbitrary_value=3731.5

        instance = MeasurementTestModel()
        instance.measurement_value=arbitrary_value
        instance.measurement_measure=arbitrary_measure_name
        instance.measurement_unit=arbitrary_default_unit
        instance.save()

        retrieved = MeasurementTestModel.objects.all()[0]

        self.assertIsInstance(
            retrieved.measurement,
            measure.UnknownMeasure
        )

    def test_storage_of_unknown_measure_stores_same_measure(self):
        arbitrary_default_unit='versta'
        arbitrary_measure_name='ImperialRussian'
        arbitrary_value=3731.5

        instance = MeasurementTestModel()
        instance.measurement_value=arbitrary_value
        instance.measurement_measure=arbitrary_measure_name
        instance.measurement_unit=arbitrary_default_unit
        instance.save()

        retrieved = MeasurementTestModel.objects.all()[0]

        new = MeasurementTestModel.objects.create(
            measurement=retrieved.measurement
        )

        first = MeasurementTestModel.objects.all()[0]
        second = MeasurementTestModel.objects.all()[1]

        self.assertEqual(
            first.measurement,
            second.measurement,
        )

class MeasurementUtilsTest(TestCase):
    def test_guess_measurement_weight(self):
        expected_measurement = measure.Weight(mcg=101)
        actual_measurement = utils.guess_measurement(
            101,
            'mcg',
        )

        self.assertEqual(
            expected_measurement,
            actual_measurement,
        )

    def test_guess_measurement_distance(self):
        expected_measurement = measure.Distance(mi=2144)
        actual_measurement = utils.guess_measurement(
            2144,
            'mi',
        )

        self.assertEqual(
            expected_measurement,
            actual_measurement,
        )

    def test_get_measurement(self):
        expected_measurement = measure.Volume(us_qt=34)
        actual_measurement = utils.get_measurement(
            measure.Volume,
            34,
            'us_qt',
        )

        self.assertEqual(
            expected_measurement,
            actual_measurement,
        )




from django_measurement.base import Distance, Area, MeasureBase, total_ordering

__all__ = [
    'Distance',
    'Area',
    'Weight',
    'Volume',
    'UnknownMeasure',
]

class Weight(MeasureBase):
    STANDARD_UNIT = 'g'
    UNITS = {
        'mcg': 0.000001,
        'mg': 0.001,
        'g': 1.0,
        'kg': 1000.0,
        'tonne': 1000000.0,
        'oz': 28.3495,
        'lb': 453.592,
        'stone': 6350.29,
        'short_ton': 907185.0,
        'long_ton': 1016000.0,
    }
    ALIAS = {
        'microgram': 'mcg',
        'milligram': 'mg',
        'gram': 'g',
        'kilogram': 'kg',
        'ton': 'short_ton',
        'metric tonne': 'tonne',
        'metric ton': 'tonne',
        'ounce': 'oz',
        'pound': 'lb',
        'short ton': 'short_ton',
        'long ton': 'long_ton',
    }
    LALIAS = dict([(k.lower(), v) for k, v in ALIAS.items()])


class Volume(MeasureBase):
    STANDARD_UNIT = 'cubic_meter'
    UNITS = {
        'us_g': 0.00378541,
        'us_qt': 0.000946353,
        'us_pint': 0.000473176,
        'us_cup': 0.000236588,
        'us_oz': 2.9574e-5,
        'us_tbsp': 1.4787e-5,
        'us_tsp': 4.9289e-6,
        'cubic_centimeter': 100.0,
        'cubic_meter': 1.0,
        'l': 0.001,
        'ml': 1e-6,
        'cubic_foot': 0.0283168,
        'cubic_inch': 1.6387e-5,
        'imperial_g': 0.00454609,
        'imperial_qt': 0.00113652,
        'imperial_pint': 0.000568261,
        'imperial_oz': 2.8413e-5,
        'imperial_tbsp': 1.7758e-5,
        'imperial_tsp': 5.9194e-6,
    }
    ALIAS = {
        'US Gallon': 'us_g',
        'US Quart': 'us_qt',
        'US Pint': 'us_pint',
        'US Cup': 'us_cup',
        'US Ounce': 'us_oz',
        'US Fluid Ounce': 'us_oz',
        'US Tablespoon': 'us_tbsp',
        'US Teaspoon': 'us_tsp',
        'cubic centimeter': 'cubic_centimeter',
        'cubic meter': 'cubic_meter',
        'liter': 'l',
        'milliliter': 'ml',
        'cubic foot': 'cubic_foot',
        'cubic inch': 'cubic_inch',
        'Imperial Gram': 'imperial_g',
        'Imperial Quart': 'imperial_qt',
        'Imperial Pint': 'imperial_pint',
        'Imperial Ounce': 'imperial_oz',
        'Imperial Tablespoon': 'imperial_tbsp',
        'Imperial Teaspoon': 'imperial_tsp',
    }
    LALIAS = dict([(k.lower(), v) for k, v in ALIAS.items()])

    def __init__(self, *args, **kwargs):
        super(Volume, self).__init__(*args, **kwargs)


@total_ordering
class UnknownMeasure(object):
    def __init__(self, measure, original_unit, value):
        self.measure = measure
        self.original_unit = original_unit
        self.value = value

    def __getattr__(self, name):
        raise AttributeError(
            'UnknownMeasures cannot be convered to other units.'
        )

    def get_measurement_parts(self):
        return self.measure, self.original_unit, self.value

    def __eq__(self, other):
        if isinstance(other, UnknownMeasure):
            if self.measure == other.measure:
                if self.value == other.value:
                    return True
        return False

    def __lt__(self, other):
        if isinstance(other, UnknownMeasure):
            if self.measure == other.measure:
                if self.value < other.value:
                    return True
                return False
        return NotImplemented

    def __repr__(self):
        return '%s(?=%s)' % (self.measure, self.value)

    def __str__(self):
        return '%s ? (%s)' % (self.value, self.measure)



from django.conf import settings
from django.db.models import signals
from django.db.models.fields import CharField, FloatField, CharField, Field

from django_measurement import measure, utils


def get_measurement_parts(value):
    if isinstance(value, measure.UnknownMeasure):
        return value.get_measurement_parts()
    measure_name = value.__class__.__name__
    original_unit = value._default_unit
    standard_value = getattr(value, value.STANDARD_UNIT)
    return measure_name, original_unit, standard_value

class MeasurementFieldDescriptor(object):
    def __init__(self, field, measurement_field_name, original_unit_field_name, measure_field_name):
        self.field = field
        self.measurement_field_name = measurement_field_name
        self.original_unit_field_name = original_unit_field_name
        self.measure_field_name = measure_field_name

    def _get_measure_by_name(self, measure_name):
        measures = utils.build_measure_list()
        return measures[measure_name]

    def __get__(self, instance, instance_type=None):
        if instance is None:
            return self

        measure_name = getattr(
            instance, 
            self.measure_field_name
        )
        measurement_value = getattr(
            instance,
            self.measurement_field_name,
        )
        original_unit = getattr(
            instance,
            self.original_unit_field_name,
        )
        try:
            instance_measure = self._get_measure_by_name(measure_name)
            measurement = utils.get_measurement(
                instance_measure,
                measurement_value,
                instance_measure.STANDARD_UNIT,
            )
            measurement._default_unit = original_unit
        except (AttributeError, KeyError, ImportError):
            measurement = measure.UnknownMeasure(
                measure=measure_name,
                original_unit=original_unit,
                value=measurement_value,
            )

        return measurement

    def __set__(self, instance, measurement):
        if instance is None:
            raise AttributeError("Must be accessed via instance")

        measure_name, original_unit, standard_value = get_measurement_parts(
            measurement
        )
        
        setattr(instance, self.measure_field_name, measure_name)
        setattr(instance, self.original_unit_field_name, original_unit)
        setattr(instance, self.measurement_field_name, standard_value)

class MeasurementField(Field):
    def __init__(self, 
            *args,
            **kwargs
        ):
        super(MeasurementField, self).__init__(*args, **kwargs)
        self.original_unit_field_name = '%s_unit'
        self.measure_field_name = '%s_measure'
        self.measurement_field_name = '%s_value'

    def get_original_unit_field_name(self):
        return self.original_unit_field_name % self.name

    def get_measure_field_name(self):
        return self.measure_field_name % self.name

    def get_measurement_field_name(self):
        return self.measurement_field_name % self.name

    def contribute_to_class(self, cls, name):
        self.name = name

        original_unit_field_name = self.get_original_unit_field_name()
        self.original_unit_field = CharField(
            editable=False,
            blank=True,
            default='',
            max_length=50,
        )
        cls.add_to_class(original_unit_field_name, self.original_unit_field)

        measure_field_name = self.get_measure_field_name()
        self.measure_field = CharField(
            editable=False,
            blank=True,
            default='',
            max_length=255,
        )
        cls.add_to_class(measure_field_name, self.measure_field)

        measurement_field_name = self.get_measurement_field_name()
        self.measurement_field = CharField(
            editable=False,
            blank=True,
            default='',
            max_length=50,
        )
        cls.add_to_class(measurement_field_name, self.measurement_field)

        field = MeasurementFieldDescriptor(
            self,
            measurement_field_name=measurement_field_name,
            original_unit_field_name=original_unit_field_name,
            measure_field_name=measure_field_name,
        )

        signals.pre_init.connect(self.instance_pre_init, sender=cls, weak=False)

        setattr(cls, self.name, field)

    def instance_pre_init(self, signal, sender, args, kwargs, **_kwargs):
        if self.name in kwargs:
            value = kwargs.pop(self.name)
            measure_name, original_unit, standard_value = get_measurement_parts(
                value
            )
            kwargs[self.get_measure_field_name()] = measure_name
            kwargs[self.get_original_unit_field_name()] = original_unit
            kwargs[self.get_measurement_field_name()] = standard_value


try:
    from south.modelsinspector import add_introspection_rules
    add_introspection_rules([], ["^django_measurement\.fields\.MeasurementField"])
except ImportError:
    pass



from decimal import Decimal

import django
from django.contrib.gis.measure import Distance, Area

try:
    from functools import total_ordering
except ImportError:
    def total_ordering(cls):
        """Class decorator that fills in missing ordering methods"""
        convert = {
            '__lt__': [('__gt__', lambda self, other: not (self < other or self == other)),
                       ('__le__', lambda self, other: self < other or self == other),
                       ('__ge__', lambda self, other: not self < other)],
            '__le__': [('__ge__', lambda self, other: not self <= other or self == other),
                       ('__lt__', lambda self, other: self <= other and not self == other),
                       ('__gt__', lambda self, other: not self <= other)],
            '__gt__': [('__lt__', lambda self, other: not (self > other or self == other)),
                       ('__ge__', lambda self, other: self > other or self == other),
                       ('__le__', lambda self, other: not self > other)],
            '__ge__': [('__le__', lambda self, other: (not self >= other) or self == other),
                       ('__gt__', lambda self, other: self >= other and not self == other),
                       ('__lt__', lambda self, other: not self >= other)]
        }
        roots = set(dir(cls)) & set(convert)
        if not roots:
            raise ValueError('must define at least one ordering operation: < > <= >=')
        root = max(roots)       # prefer __lt__ to __le__ to __gt__ to __ge__
        for opname, opfunc in convert[root]:
            if opname not in roots:
                opfunc.__name__ = opname
                opfunc.__doc__ = getattr(int, opname).__doc__
                setattr(cls, opname, opfunc)
        return cls

if django.VERSION[0] > 1 or (django.VERSION[0] == 1 and django.VERSION[1] > 4):
    from django.contrib.gis.measure import MeasureBase
else:
    NUMERIC_TYPES = (int, long, float, Decimal)
    # Set STANDARD_UNIT for existing Distance/Area classes
    Distance.STANDARD_UNIT = 'm'
    Area.STANDARD_UNIT = 'sq_m'

    def pretty_name(obj):
        return obj.__name__ if obj.__class__ == type else obj.__class__.__name__

    @total_ordering
    class MeasureBase(object):
        STANDARD_UNIT = None
        ALIAS  = {}
        UNITS  = {}
        LALIAS = {}

        def __init__(self, default_unit=None, **kwargs):
            value, self._default_unit = self.default_units(kwargs)
            setattr(self, self.STANDARD_UNIT, value)
            if default_unit and isinstance(default_unit, basestring):
                self._default_unit = default_unit

        def _get_standard(self):
            return getattr(self, self.STANDARD_UNIT)

        def _set_standard(self, value):
            setattr(self, self.STANDARD_UNIT, value)

        standard = property(_get_standard, _set_standard)

        def __getattr__(self, name):
            if name in self.UNITS:
                return self.standard / self.UNITS[name]
            else:
                raise AttributeError('Unknown unit type: %s' % name)

        def __repr__(self):
            return '%s(%s=%s)' % (pretty_name(self), self._default_unit,
                getattr(self, self._default_unit))

        def __str__(self):
            return '%s %s' % (getattr(self, self._default_unit), self._default_unit)

        # **** Comparison methods ****

        def __eq__(self, other):
            if isinstance(other, self.__class__):
                return self.standard == other.standard
            else:
                return NotImplemented

        def __lt__(self, other):
            if isinstance(other, self.__class__):
                return self.standard < other.standard
            else:
                return NotImplemented

        # **** Operators methods ****

        def __add__(self, other):
            if isinstance(other, self.__class__):
                return self.__class__(default_unit=self._default_unit,
                    **{self.STANDARD_UNIT: (self.standard + other.standard)})
            else:
                raise TypeError('%(class)s must be added with %(class)s' % {"class":pretty_name(self)})

        def __iadd__(self, other):
            if isinstance(other, self.__class__):
                self.standard += other.standard
                return self
            else:
                raise TypeError('%(class)s must be added with %(class)s' % {"class":pretty_name(self)})

        def __sub__(self, other):
            if isinstance(other, self.__class__):
                return self.__class__(default_unit=self._default_unit,
                    **{self.STANDARD_UNIT: (self.standard - other.standard)})
            else:
                raise TypeError('%(class)s must be subtracted from %(class)s' % {"class":pretty_name(self)})

        def __isub__(self, other):
            if isinstance(other, self.__class__):
                self.standard -= other.standard
                return self
            else:
                raise TypeError('%(class)s must be subtracted from %(class)s' % {"class":pretty_name(self)})

        def __mul__(self, other):
            if isinstance(other, NUMERIC_TYPES):
                return self.__class__(default_unit=self._default_unit,
                    **{self.STANDARD_UNIT: (self.standard * other)})
            else:
                raise TypeError('%(class)s must be multiplied with number' % {"class":pretty_name(self)})

        def __imul__(self, other):
            if isinstance(other, NUMERIC_TYPES):
                self.standard *= float(other)
                return self
            else:
                raise TypeError('%(class)s must be multiplied with number' % {"class":pretty_name(self)})

        def __rmul__(self, other):
            return self * other

        def __truediv__(self, other):
            if isinstance(other, self.__class__):
                return self.standard / other.standard
            if isinstance(other, NUMERIC_TYPES):
                return self.__class__(default_unit=self._default_unit,
                    **{self.STANDARD_UNIT: (self.standard / other)})
            else:
                raise TypeError('%(class)s must be divided with number or %(class)s' % {"class":pretty_name(self)})

        def __div__(self, other):   # Python 2 compatibility
            return type(self).__truediv__(self, other)

        def __itruediv__(self, other):
            if isinstance(other, NUMERIC_TYPES):
                self.standard /= float(other)
                return self
            else:
                raise TypeError('%(class)s must be divided with number' % {"class":pretty_name(self)})

        def __idiv__(self, other):  # Python 2 compatibility
            return type(self).__itruediv__(self, other)

        def __bool__(self):
            return bool(self.standard)

        def __nonzero__(self):      # Python 2 compatibility
            return type(self).__bool__(self)

        def default_units(self, kwargs):
            """
            Return the unit value and the default units specified
            from the given keyword arguments dictionary.
            """
            val = 0.0
            default_unit = self.STANDARD_UNIT
            for unit, value in kwargs.items():
                if not isinstance(value, float): value = float(value)
                if unit in self.UNITS:
                    val += self.UNITS[unit] * value
                    default_unit = unit
                elif unit in self.ALIAS:
                    u = self.ALIAS[unit]
                    val += self.UNITS[u] * value
                    default_unit = u
                else:
                    lower = unit.lower()
                    if lower in self.UNITS:
                        val += self.UNITS[lower] * value
                        default_unit = lower
                    elif lower in self.LALIAS:
                        u = self.LALIAS[lower]
                        val += self.UNITS[u] * value
                        default_unit = u
                    else:
                        raise AttributeError('Unknown unit type: %s' % unit)
            return val, default_unit

        @classmethod
        def unit_attname(cls, unit_str):
            """
            Retrieves the unit attribute name for the given unit string.
            For example, if the given unit string is 'metre', 'm' would be returned.
            An exception is raised if an attribute cannot be found.
            """
            lower = unit_str.lower()
            if unit_str in cls.UNITS:
                return unit_str
            elif lower in cls.UNITS:
                return lower
            elif lower in cls.LALIAS:
                return cls.LALIAS[lower]
            else:
                raise Exception('Could not find a unit keyword associated with "%s"' % unit_str)





import inspect

from django.conf import settings


MEASURE_OVERRIDES = getattr(settings, 'MEASURE_OVERRIDES', {})


def get_class_by_path(path):
    mod = __import__('.'.join(path.split('.')[:-1]))
    components = path.split('.')
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod


def build_measure_list():
    from django.contrib.gis.measure import MeasureBase as DjangoMeasureBase
    from django_measurement.base import MeasureBase
    from django_measurement import measure
    measures = {}
    possible_measures = dir(measure)
    for possible_measure_name in possible_measures:
        possible_measure = getattr(measure, possible_measure_name)
        if not inspect.isclass(possible_measure):
            continue
        if issubclass(possible_measure, (MeasureBase, DjangoMeasureBase, )):
            measures[possible_measure_name] = possible_measure
    for overridden_measure_name, cls_path in MEASURE_OVERRIDES.items():
        cls = get_class_by_path(
            cls_path
        )
        measures[overridden_measure_name] = cls
    return measures


def get_measurement(measure, value, unit):
    return measure(
        **{unit: value}
    )
    

def guess_measurement(value, unit):
    all_measures = build_measure_list()
    for measure_name, measure in all_measures.items():
        all_units = getattr(measure, 'UNITS', {}).keys()
        if unit in all_units:
            return get_measurement(measure, value, unit)



from setuptools import setup

tests_require=['django']

setup(
    name='django-measurement',
    version='0.9',
    url='http://bitbucket.org/latestrevision/django-measurement/',
    description='Convenient fields and classes for handling measurements',
    author='Adam Coddington',
    author_email='me@adamcoddington.net',
    tests_require=tests_require,
    extras_require={'test': tests_require},
    test_suite='django_measurement.runtests.runtests',
    classifiers=[
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],
    packages=[
        'django_measurement',
        ],
)
